package schema

import (
	"context"

	"entgo.io/contrib/entgql"
	"entgo.io/ent"
	"entgo.io/ent/dialect/entsql"
	"entgo.io/ent/schema"
	"entgo.io/ent/schema/edge"
	"entgo.io/ent/schema/field"
	"entgo.io/ent/schema/index"
	gen "github.com/valocode/bubbly/ent"
	"github.com/valocode/bubbly/ent/extensions/entmodel"
	"github.com/valocode/bubbly/ent/hook"
	types "github.com/valocode/bubbly/ent/schema/types"
	"github.com/valocode/bubbly/ent/vulnerability"
)

// TODO: How to handle aliases
type Vulnerability struct {
	ent.Schema
}

func (Vulnerability) Annotations() []schema.Annotation {
	return []schema.Annotation{
		entsql.Annotation{Table: "vulnerability"},
		entmodel.Annotation{},
	}
}

func (Vulnerability) Fields() []ent.Field {
	return []ent.Field{
		field.String("vid").
			NotEmpty().
			Unique().
			Immutable().
			Annotations(
				entgql.OrderField("vid"),
			),
		field.String("summary").Optional(),
		field.String("description").Optional(),
		field.Float("severity_score").
			Default(0.0).
			Annotations(
				entgql.OrderField("severity_score"),
			),
		field.Enum("severity").
			Default("None").
			Values("None", "Low", "Medium", "High", "Critical").
			Annotations(
				entgql.OrderField("severity"),
			),
		field.Time("published").Optional().
			Annotations(
				entgql.OrderField("published"),
			),
		field.Time("modified").Optional().
			Annotations(
				entgql.OrderField("modified"),
			),
		field.JSON("metadata", types.Metadata{}).Optional(),
	}
}

func (Vulnerability) Edges() []ent.Edge {
	return []ent.Edge{
		edge.To("owner", Organization.Type).Unique().Required(),
		edge.From("components", Component.Type).Ref("vulnerabilities"),
		edge.From("reviews", VulnerabilityReview.Type).Ref("vulnerability"),
		edge.From("instances", ReleaseVulnerability.Type).Ref("vulnerability"),
	}
}

func (Vulnerability) Indexes() []ent.Index {
	return []ent.Index{
		index.Fields("vid").Unique(),
	}
}

func (Vulnerability) Hooks() []ent.Hook {
	return []ent.Hook{
		// Automatically set the Vulnerability severity if it has not already been set
		hook.On(
			func(next ent.Mutator) ent.Mutator {
				return hook.VulnerabilityFunc(func(ctx context.Context, m *gen.VulnerabilityMutation) (ent.Value, error) {
					if score, exists := m.SeverityScore(); exists {
						// Ranges: https://nvd.nist.gov/vuln-metrics/cvss
						var severity vulnerability.Severity
						switch {
						case score == 0.0:
							severity = vulnerability.SeverityNone
						case score >= 0.1 && score < 4.0:
							severity = vulnerability.SeverityLow
						case score >= 4.0 && score < 7.0:
							severity = vulnerability.SeverityMedium
						case score >= 7.0 && score < 9.0:
							severity = vulnerability.SeverityHigh
						case score >= 9.0:
							severity = vulnerability.SeverityCritical
						default:
							severity = vulnerability.SeverityNone
						}
						m.SetSeverity(severity)
					}
					return next.Mutate(ctx, m)
				})
			},
			ent.OpCreate|ent.OpUpdate|ent.OpUpdateOne,
		),
	}
}
